@@ -1,3 +1,4 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 2 -*- */
 /* pager object */
 
 /*
@@ -28,7 +29,7 @@
 #include "xutils.h"
 #include "pager-accessible-factory.h"
 #include "workspace-accessible-factory.h"
-
+#include "private.h"
 
 #define N_SCREEN_CONNECTIONS 11
 
@@ -59,6 +60,10 @@ struct _WnckPagerPrivate
   GdkPixbuf *bg_cache;
 
   int layout_manager_token;
+
+  guint dnd_activate;
+  gint  dnd_workspace_number;
+  guint dnd_time;
 };
 
 enum
@@ -88,6 +93,14 @@ static gboolean wnck_pager_expose_event  (GtkWidget        *widget,
                                           GdkEventExpose   *event);
 static gboolean wnck_pager_button_press  (GtkWidget        *widget,
                                           GdkEventButton   *event);
+static gboolean wnck_pager_drag_motion   (GtkWidget        *widget,
+                                          GdkDragContext   *context,
+                                          gint              x,
+                                          gint              y,
+                                          guint             time);
+static void wnck_pager_drag_motion_leave (GtkWidget        *widget,
+                                          GdkDragContext   *context,
+                                          guint             time);
 static gboolean wnck_pager_motion        (GtkWidget        *widget,
                                           GdkEventMotion   *event);
 static gboolean wnck_pager_button_release (GtkWidget        *widget,
@@ -115,7 +128,6 @@ static AtkObject* wnck_pager_get_accessible (GtkWidget *widget);
 
 
 static gpointer parent_class;
-static guint signals[LAST_SIGNAL] = { 0 };
 
 GType
 wnck_pager_get_type (void)
@@ -161,6 +173,8 @@ wnck_pager_init (WnckPager *pager)
   pager->priv->bg_cache = NULL;
   pager->priv->layout_manager_token = WNCK_NO_MANAGER_TOKEN;
 
+  gtk_drag_dest_set (GTK_WIDGET (pager), 0, NULL, 0, 0);
+
   GTK_WIDGET_SET_FLAGS (GTK_WIDGET (pager), GTK_CAN_FOCUS);
 }
 
@@ -183,7 +197,9 @@ wnck_pager_class_init (WnckPagerClass *klass)
   widget_class->button_release_event = wnck_pager_button_release;
   widget_class->motion_notify_event = wnck_pager_motion;
   widget_class->focus = wnck_pager_focus;
-  widget_class->get_accessible = wnck_pager_get_accessible;   
+  widget_class->get_accessible = wnck_pager_get_accessible;
+  widget_class->drag_leave = wnck_pager_drag_motion_leave;
+  widget_class->drag_motion = wnck_pager_drag_motion;	
 }
 
 static void
@@ -200,7 +216,13 @@ wnck_pager_finalize (GObject *object)
       g_object_unref (G_OBJECT (pager->priv->bg_cache));
       pager->priv->bg_cache = NULL;
     }
-  
+
+  if (pager->priv->dnd_activate != 0)
+    {
+      g_source_remove (pager->priv->dnd_activate);
+      pager->priv->dnd_activate = 0;
+    }
+
   g_free (pager->priv);
   
   G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -545,6 +567,8 @@ get_windows_for_workspace_in_bottom_to_top (WnckScreen    *screen,
         continue;
       if (!is_active && wnck_window_is_sticky (win))
         continue;
+      if (!is_active && wnck_window_is_pinned (win))
+        continue;
 
       result = g_list_prepend (result, win);
     }
@@ -594,14 +618,15 @@ get_window_rect (WnckWindow         *window,
   unclipped_win_rect.width = width;
   unclipped_win_rect.height = height;
 
-  gdk_rectangle_intersect (workspace_rect, &unclipped_win_rect, rect);
+  gdk_rectangle_intersect ((GdkRectangle *) workspace_rect, &unclipped_win_rect, rect);
 }
 
 static void
 draw_window (GdkDrawable        *drawable,
              GtkWidget          *widget,
              WnckWindow         *win,
-             const GdkRectangle *winrect)
+             const GdkRectangle *winrect,
+             gboolean            on_current_workspace)
 {
   GdkPixbuf *icon;
   int icon_x, icon_y, icon_w, icon_h;
@@ -610,7 +635,7 @@ draw_window (GdkDrawable        *drawable,
   is_active = wnck_window_is_active (win);
           
   gdk_draw_rectangle (drawable,
-                      is_active ?
+                      is_active && on_current_workspace ?
                       widget->style->bg_gc[GTK_STATE_SELECTED] :
                       widget->style->bg_gc[GTK_STATE_NORMAL],
                       TRUE,
@@ -863,7 +888,8 @@ wnck_pager_draw_workspace (WnckPager    *pager,
 	  draw_window (GTK_WIDGET (pager)->window,
 		       GTK_WIDGET (pager),
 		       win,
-		       &winrect);
+		       &winrect,
+                       is_current);
 	  
 	  tmp = tmp->next;
 	}
@@ -990,6 +1016,13 @@ wnck_pager_expose_event  (GtkWidget      *widget,
       GdkRectangle rect;
       GdkRectangle winrect;
       int dx, dy;
+      WnckWorkspace *active_space;
+      gboolean is_current;
+  
+      active_space = wnck_screen_get_active_workspace (pager->priv->screen);
+
+      is_current = active_space &&
+        drag_space == wnck_workspace_get_number (active_space);
       
       get_workspace_rect (pager, drag_space, &rect);          
       get_window_rect (pager->priv->drag_window, &rect, &winrect);
@@ -1005,7 +1038,8 @@ wnck_pager_expose_event  (GtkWidget      *widget,
       draw_window (widget->window,
                    widget,
                    pager->priv->drag_window,
-                   &winrect);
+                   &winrect,
+                   is_current);
     }
   
   return FALSE;
@@ -1080,6 +1114,65 @@ wnck_pager_button_press (GtkWidget      *widget,
   return handled;
 }
 
+static gboolean
+wnck_pager_drag_motion_timeout (gpointer data)
+{
+  WnckPager *pager = WNCK_PAGER (data);
+  WnckWorkspace *active_workspace, *dnd_workspace;
+
+  pager->priv->dnd_activate = 0;
+  active_workspace = wnck_screen_get_active_workspace (pager->priv->screen);
+  dnd_workspace    = wnck_screen_get_workspace (pager->priv->screen,
+                                                pager->priv->dnd_workspace_number);
+
+  if (dnd_workspace &&
+      (pager->priv->dnd_workspace_number != wnck_workspace_get_number (active_workspace)))
+    wnck_workspace_activate (dnd_workspace, pager->priv->dnd_time);
+
+  return FALSE;
+}
+
+static gboolean 
+wnck_pager_drag_motion (GtkWidget          *widget,
+                        GdkDragContext     *context,
+                        gint                x,
+                        gint                y,
+                        guint               time)
+{
+  WnckPager *pager;
+
+  pager = WNCK_PAGER (widget);
+
+  if (pager->priv->dnd_activate == 0 )
+    pager->priv->dnd_activate = g_timeout_add (WNCK_ACTIVATE_TIMEOUT,
+                                               wnck_pager_drag_motion_timeout,
+                                               pager);
+
+  pager->priv->dnd_workspace_number = workspace_at_point (pager, x, y, NULL, NULL);
+  pager->priv->dnd_time = time;
+  gdk_drag_status (context, 0, time);
+
+  return (pager->priv->dnd_workspace_number != -1);
+}
+ 
+static void
+wnck_pager_drag_motion_leave (GtkWidget          *widget,
+                              GdkDragContext     *context,
+                              guint               time)
+{
+  WnckPager *pager;
+
+  pager = WNCK_PAGER (widget);
+
+  if (pager->priv->dnd_activate != 0)
+    {
+      g_source_remove (pager->priv->dnd_activate);
+      pager->priv->dnd_workspace_number = -1;
+      pager->priv->dnd_time = 0;
+      pager->priv->dnd_activate = 0;
+    }
+}
+
 static gboolean
 wnck_pager_motion (GtkWidget        *widget,
                    GdkEventMotion   *event)
@@ -1135,7 +1228,7 @@ wnck_pager_button_release (GtkWidget        *widget,
               wnck_window_move_to_workspace (pager->priv->drag_window,
                                              space);
               if (space == wnck_screen_get_active_workspace (pager->priv->screen))
-                wnck_window_activate (pager->priv->drag_window);
+                wnck_window_activate (pager->priv->drag_window, event->time);
             }	  
 	}
       
@@ -1152,7 +1245,7 @@ wnck_pager_button_release (GtkWidget        *widget,
 
 	    /* Don't switch the desktop if we're already there */
 	    if (space != wnck_screen_get_active_workspace (pager->priv->screen))
-	      wnck_workspace_activate (space);
+	      wnck_workspace_activate (space, event->time);
 
 	    /* EWMH only lets us move the viewport for the active workspace,
 	     * but we just go ahead and hackily assume that the activate
@@ -1171,14 +1264,6 @@ wnck_pager_button_release (GtkWidget        *widget,
 	    if (wnck_workspace_get_viewport_x (space) != viewport_x ||
 		wnck_workspace_get_viewport_y (space) != viewport_y)
 	      wnck_screen_move_viewport (pager->priv->screen, viewport_x, viewport_y);
-            /* Activate the window clicked on *only* if it is in the
-             * current workspace.  Otherwise, the default window
-             * should be activated (but that's a job for the WM, not
-             * for the pager, as it is focus mode dependent).
-             */
-	    if ((space == wnck_screen_get_active_workspace (pager->priv->screen))
-                && (pager->priv->drag_window))
-	      wnck_window_activate (pager->priv->drag_window);
 	  }
       
       if (pager->priv->drag_window)
@@ -1777,9 +1862,10 @@ _wnck_pager_get_workspace (WnckPager *pager,
 } 
 
 void 
-_wnck_pager_activate_workspace (WnckWorkspace *wspace)
+_wnck_pager_activate_workspace (WnckWorkspace *wspace,
+                                guint32        timestamp)
 {
-  wnck_workspace_activate (wspace);
+  wnck_workspace_activate (wspace, timestamp);
 }
 
 void
